""" Contains all the built-in :class:`.Arbitrary` instances. They all can be
imported with ``from pyquchk.arbitraries import ...``, no need to specify
exact modules.
"""
from .primitives import *
from .numbers import *
from .sequences import *
